import numpy as np
import time
import scipy.linalg
from .activator.hgm import batch_data_loader as batch_dl

class ntk:
    def __init__(self,layers,activator,beta=0):
        self.activator = activator
        self.layers = layers
        self.beta = beta
        self.pred_mode = False
       
    def ev_kernel(self,x1,x2,mode,rtol = 1e-10,atol = 1e-10 ):
        def gauss_herm(f,covar,q = 50):
            a = np.sqrt(covar[0])
            b = np.sqrt(covar[2])
            c = np.min([covar[1]/(a*b),1])
            
            k_sigma = 0
            x , w = scipy.special.roots_hermite(q)
            for i in range(q):
                for j in range(q):
                    x_i = x[i]
                    w_i = w[i]
                    x_j = x[j]
                    w_j = w[j]
            
                    k_sigma += w_i*w_j*(f(np.sqrt(2)*a*x_i)*f(np.sqrt(2)*b*c*x_i + np.sqrt(2)*b*np.sqrt(1-c**2)*x_j))

            return k_sigma/np.pi
        debug_mode =False
        var1 = np.linalg.norm(x1)**2 + self.beta**2
        var2 = np.linalg.norm(x2)**2 + self.beta**2
        if x1.shape[0] > 1:
            covar_x1_x2 = x1.T@x2 + self.beta**2
        elif x1.shape[0] == 1 or x1.shape[0] == 0:
            covar_x1_x2 = x1.item()*x2.item()+self.beta**2
        sigma_l = [covar_x1_x2]
        sigma_dot_l = []
        
        for i in range(self.layers-1):
            c1 = np.sqrt(var1)
            c2 = np.sqrt(var2)
            if c1 != 0 and c2 != 0:
                r = sigma_l[i]/(c1*c2)

            else:
                r = sigma_l[i]
            # print()
            covar_matrix = np.array(([c1**2,c1*c2*r],[c1*c2*r,c2**2]))
            var1,var2, sigma ,sigma_dot = self.activator.ev_sigmas(covar_matrix,mode,rtol,atol)
            
            if debug_mode == True:
                if x1 == 1 and x2 ==-1:
                    print(covar_matrix,sigma,sigma_dot)
                covar = covar_matrix
                tmp = gauss_herm(self.activator.activator,[covar[0][0],covar[0][1],covar[1][1]])
                tmp_diff = gauss_herm(self.activator.activator_dot,[covar[0][0],covar[0][1],covar[1][1]])
                print('gauss-herm:',tmp)
                print('re:',np.abs((self.activator.c_sigma*tmp - sigma)/tmp))
                if np.abs((self.activator.c_sigma*tmp - sigma)/tmp) >=0.2:
                    print(' error:',covar_matrix,self.activator.c_sigma*tmp,sigma)
                if np.abs((self.activator.c_sigma*tmp_diff - sigma_dot)/tmp_diff) >=0.2:
                    print('diff error:',covar_matrix,self.activator.c_sigma*tmp_diff ,sigma_dot)

            if np.abs(sigma**2) > np.abs(var1*var2):
                print(f'correct sigma: x = ({x1,x2}), layer {i}')
                sigma = sigma*np.sqrt(var1*var2)/np.abs(sigma)
            #print(f'sigma:{sigma,sigma_dot}')
            sigma = sigma+self.beta**2
            sigma_l = np.append(sigma_l, sigma)
            sigma_dot_l = np.append(sigma_dot_l, sigma_dot)
            var1 = var1+self.beta**2
            var2 = var2+self.beta**2
        #print('sigmas',sigma_l.shape)
        ntk = 0
        
        for i in range(self.layers):
            #print(i,sigma_l[i])
            sigma_dot = 1
            for j in range(i+1,self.layers):
                #print(sigma_dot_l[j-1],j)
                sigma_dot = sigma_dot*sigma_dot_l[j-1]
            #print()
            #print(sigma_l[i])
            ntk = ntk+(sigma_dot*sigma_l[i])
            #print(ntk)
        if x1 == 1 and x2 ==-1:
            print(x1,x2,ntk,sigma_l,sigma_dot_l)
        
        return ntk

            
        
    def train(self,sample_x,sample_y,diag_reg = 0.01,mode = 'hgm',get_timing = False,rtol = 1e-10,atol = 1e-10):
        self.sample_x = sample_x
        self.sample_y = sample_y
        self.pred_mode = False
        self.kernel_mat = np.zeros((self.sample_x.shape[0],self.sample_x.shape[0]))
        if get_timing == True:
           start = time.time()  
        for i in range(self.sample_x.shape[0]):
            #for j in range(i+1):
            for j in range(self.sample_x.shape[0]):
                x1 = sample_x[i]
                x2 = sample_x[j]
                # print('x:',x1,x2)
                ntk = self.ev_kernel(x1,x2,mode,rtol,atol)
                self.kernel_mat[i][j] = ntk
                self.kernel_mat[j][i] = ntk
        if get_timing == True:
           end = time.time()
           training_time = end - start
           start = time.time()
        self.pred_matrix = scipy.linalg.solve(self.kernel_mat+diag_reg*np.identity(self.sample_x.shape[0]),self.sample_y,assume_a = 'sym')
        if get_timing == True:
            end = time.time()
            solve_time = end-start
        self.pred_mode = True
        print('Training finished.')
        if get_timing == True:
            print(f'Training time:{training_time}, solve_time: {solve_time}')
        

        
    def pred(self,x,mode = 'hgm',debug_mode = False,rtol = 1e-10,atol = 1e-10):
        output = []
        start = time.time()  
        for i in range(x.shape[0]):
            input_kernel = np.zeros(self.sample_x.shape[0])
            for j in range(self.sample_x.shape[0]):
                input_kernel[j] = self.ev_kernel(x[i],self.sample_x[j],mode,rtol,atol)
            if debug_mode == True:
                print(input_kernel)
            
            if self.pred_mode == True:
                kernel_output = input_kernel@self.pred_matrix
                output = np.append(output,kernel_output)
            else:
                print('You must have finished ntk.train() before running ntk.pred()')
                break
        end = time.time()
        print('pred time:',end-start)
        return output
    
    def hgm_fast_ev_kernel(self,eval_pts, rtol = 1e-10,atol = 1e-10,debug_mode = False):
        def gauss_herm(f,covar,q = 50):
            a = np.sqrt(covar[0])
            b = np.sqrt(covar[2])
            c = np.min([covar[1]/(a*b),1])
            
            k_sigma = 0
            x , w = scipy.special.roots_hermite(q)
            for i in range(q):
                for j in range(q):
                    x_i = x[i]
                    w_i = w[i]
                    x_j = x[j]
                    w_j = w[j]
            
                    k_sigma += w_i*w_j*(f(np.sqrt(2)*a*x_i)*f(np.sqrt(2)*b*c*x_i + np.sqrt(2)*b*np.sqrt(1-c**2)*x_j))

            return k_sigma/np.pi
        dim = int(eval_pts.shape[1]/2)
        sigma_array = np.zeros((self.layers,eval_pts.shape[0]))
        sigma_dot_array = np.zeros((self.layers,eval_pts.shape[0]))
        c1_list = np.zeros((self.layers,eval_pts.shape[0]))
        c2_list = np.zeros((self.layers,eval_pts.shape[0]))
        for l in range(self.layers-1):
            nc2_list = []
            degene_covars = []
            dat = []
            
            for i in range(eval_pts.shape[0]):
                if l == 0:
                    if dim == 1 or dim == 0:
                        c1 = eval_pts[i][0].item()**2 + self.beta**2
                        c2 = eval_pts[i][1].item()**2 + self.beta**2
                        sigma = eval_pts[i][0].item()*eval_pts[i][1].item() + self.beta**2
                        #c1_list[l][i] = c1
                        #c2_list[l][i] = c2
                        sigma_array[l][i] = sigma

                if l > 0:
                    c1 = c1_list[l-1][i] 
                    c2 = c2_list[l-1][i]
                    sigma = sigma_array[l][i]
                if l+1 <= self.layers-1:
                    # 第2層以降のc1,c2
                    a = np.sqrt(c1)
                    b = np.sqrt(c2)
                    var1 = 0
                    var2 = 0
                    q = 50
                    x , w = scipy.special.roots_hermite(q)
                    for s in range(q):
                        for t in range(q):
                            x_i = x[s]
                            w_i = w[s]
                            x_j = x[t]
                            w_j = w[t]
                            var1 += w_i*w_j*(self.activator(np.sqrt(2)*a*x_i)*self.activator(np.sqrt(2)*a*1*x_i))
                            var2 += w_i*w_j*(self.activator(np.sqrt(2)*b*x_i)*self.activator(np.sqrt(2)*b*1*x_i))
                    var1 /= np.pi
                    var2 /= np.pi
                    c1_list[l][i] = self.activator.c_sigma*var1+self.beta**2
                    c2_list[l][i] = self.activator.c_sigma*var2+self.beta**2
                covar = covar = np.array([[c1,sigma],[sigma,c2]])
                #print('covar:',covar)
                det = np.linalg.det(covar)
                # print('det',det)
                if det > 1e-3:
                    mat_inv =(-1/2)* np.linalg.inv(covar)
                    #print(mat_inv)
                    covar_inv =np.array([mat_inv[0][0],mat_inv[0][1],mat_inv[1][1],eval_pts[i][-1]])
                    dat = np.append(dat,covar_inv)

                else:
                    degene_covars = np.append(degene_covars,np.array([covar[0][0],covar[0][1],covar[1][1],eval_pts[i][-1]]))

            degene_covars = np.array(degene_covars).reshape(-1,4)
            #print('degene_covars =',degene_covars)
            dat = dat.reshape(-1,4)

            [dat2,removed,removed_small_covar]=batch_dl.sort_dat(dat,small_covar=self.activator.hgm_small_covar,large_covar = self.activator.hgm_large_covar,threshould=1e-5,initial_point = self.activator.batch_init,even = None)
            # [dat2,removed,removed_small_covar]=batch_dl.sort_dat(dat,small_covar=1e-3,threshould=1e-5,initial_point = np.array([[-1.0,1/100,-1.0, -1]]),even = None)
            removed = removed.reshape(-1,5)
            dat2 = np.vstack((dat2,dat2[0]))

            # print('removed:',removed)
            # print('removed_small_covar:',removed_small_covar)
            # if l == 0 and debug_mode == True:
            #     print('dat2:',dat2)
            #     np.savetxt('2024-10-20.csv',dat2)
            #     print('saved')
            sol = self.activator.batch_hgm(dat2,qsize=-1,rtol = rtol,atol = atol)
            # 最後の点の解を求める
            # tmp_sol = self.activator._hgm_sigma(dat2[-1])
            # print(tmp_sol)

            hgm_sol = sol[0].reshape(-1,1)
            #hgm_sol = np.vstack((hgm_sol,[[tmp_sol[0][-1]]]))

            diff_sol = self.activator.batch_hgm_dot(dat2,qsize=-1,rtol = rtol,atol = atol)
            hgm_diff_sol = diff_sol[0].reshape(-1,1)
            # tmp_diff_sol = self.activator._hgm_sigma_dot(dat2[-1])
            # print(tmp_diff_sol)
            # hgm_diff_sol = np.vstack((hgm_diff_sol,[tmp_diff_sol[0,-1]]))

            gh_sol = []
            gh_diff_sol = [] #微分の期待値
            for i in range(degene_covars.shape[0]):
                tmp = gauss_herm(self.activator.activator,degene_covars[i][0:3])
                tmp_diff = gauss_herm(self.activator.activator_dot,degene_covars[i][0:3])
                
                gh_diff_sol = np.append(gh_diff_sol,tmp_diff)
                gh_sol = np.append(gh_sol,tmp)
            small_covar_sol = []
            small_covar_diff_sol =[]
            for i in range(removed_small_covar.shape[0]):
                covar = np.linalg.inv(-2*np.array([[removed_small_covar[i][0],removed_small_covar[i][1]],[removed_small_covar[i][1],removed_small_covar[i][2]]]))
                print('covar_mat:',covar)
                # tmp = self.activator._adaptive_gauss_herm(self.activator.activator,[covar[0][0],covar[0][1],covar[1][1]],accuracy_dig= rtol)
                # tmp_diff = self.activator._adaptive_gauss_herm(self.activator.activator_dot,[covar[0][0],covar[0][1],covar[1][1]],accuracy_dig= rtol)
                _,_,tmp1,tmp2 = self.activator._gauss_herm_sigmas(covar,accuracy_dig = rtol)
                small_covar_sol = np.append(small_covar_sol,tmp1)
                small_covar_diff_sol =np.append(small_covar_diff_sol,tmp2)

                

            for i in range(dat2.shape[0]-1):
                idx =int(dat2[i][3])
                x = np.array([[dat2[i][0],dat2[i][1]],[dat2[i][1],dat2[i][2]]])
                covar = np.linalg.inv(-2*x)
                nc2x = self.activator.nc2(dat2[i][0:3])
                if debug_mode == True:
                    tmp = gauss_herm(self.activator.activator,[covar[0][0],covar[0][1],covar[1][1]])
                    tmp_diff = gauss_herm(self.activator.activator_dot,[covar[0][0],covar[0][1],covar[1][1]])
                    print('tmp:',tmp)

                    if np.abs((tmp - hgm_sol[i]/nc2x)/tmp) >=0.2:
                        tmp_sol = self.activator._hgm_sigma(dat2[i])
                        print(' error:',eval_pts[idx],tmp ,hgm_sol[i]/nc2x,tmp_sol,np.abs((tmp - hgm_sol[i]/nc2x)/tmp))

                    if np.abs((tmp_diff - hgm_diff_sol[i]/nc2x)/tmp) >=0.2:

                        print('diff error:',eval_pts[idx],tmp_diff ,hgm_diff_sol[i],np.abs((tmp_diff - hgm_diff_sol[i]/nc2x)/tmp))
                    
                if idx >= 0:
                    if l < self.layers-1:
                        #print(hgm_sol[i],hgm_diff_sol[i])
                        sigma_array[l+1][idx] = (self.activator.c_sigma*hgm_sol[i])/nc2x+self.beta**2
                        sigma_dot_array[l][idx] = (self.activator.c_sigma*hgm_diff_sol[i])/nc2x
            for i in range(removed.shape[0]):
                idx1 = int(removed[i][3])
                idx2 = int(removed[i][4])
                if l < self.layers-1:
                    # gauss-hermの値と比較する
                    sigma_array[l+1][idx1] = sigma_array[l+1][idx2]
                    sigma_dot_array[l][idx1] = sigma_dot_array[l][idx2]
                    
            for i in range(degene_covars.shape[0]):
                idx = int(degene_covars[i][3])
                if l < self.layers-1:
                    sigma_array[l+1][idx] = self.activator.c_sigma*gh_sol[i]+self.beta**2
                    sigma_dot_array[l][idx] = self.activator.c_sigma*gh_diff_sol[i]
            for i in range(removed_small_covar.shape[0]):
                idx = int(removed_small_covar[i][3])
                # covar = np.linalg.inv(-2*np.array([[removed_small_covar[i][0],removed_small_covar[i][1]],[removed_small_covar[i][1],removed_small_covar[i][2]]]))
                # _,_,tmp,tmp2 = self.activator._gauss_herm_sigmas(covar,accuracy_dig = rtol)
                # print('small_covar:',small_covar_sol[i],small_covar_diff_sol[i])
                # print('gh:',tmp,tmp2)
                if l < self.layers-1:
                    sigma_array[l+1][idx] = self.activator.c_sigma*small_covar_sol[i]+self.beta**2
                    sigma_dot_array[l][idx] = self.activator.c_sigma*small_covar_diff_sol[i]
        return sigma_array,sigma_dot_array
    
    def hgm_fast_train(self,sample_x,sample_y,diag_reg = 0.01,rtol = 1e-10,atol = 1e-10,mode = 'hgm',get_timing = False,debug_mode = False):
        self.sample_x = sample_x
        self.sample_y = sample_y
        self.pred_mode = False
        self.kernel_mat = np.zeros((self.sample_x.shape[0]*self.sample_x.shape[0]))
        if get_timing == True:
           start = time.time()  
        
        dim = self.sample_x.shape[1]
        sample_pts = batch_dl.make_pair(self.sample_x)
        sample_pts = batch_dl.add_index(sample_pts)
        sigma_array, sigma_dot_array = self.hgm_fast_ev_kernel(sample_pts,rtol = rtol, atol = atol,debug_mode = debug_mode)
            
        for i in range(sample_pts.shape[0]):
            ntk = 0
            for s in range(self.layers):

                sigma_dot = 1
                for t in range(s+1,self.layers):
                    #print(sigma_dot_array[t-1][i])
                    
                    sigma_dot = sigma_dot*sigma_dot_array[t-1][i]
                
                ntk = ntk +(sigma_dot*sigma_array[s][i])
            self.kernel_mat[i] = ntk

        self.kernel_mat = self.kernel_mat.reshape(self.sample_x.shape[0],-1)
        if get_timing == True:
           end = time.time()
           training_time = end - start
           start = time.time()
        self.pred_matrix = scipy.linalg.solve(self.kernel_mat+diag_reg*np.identity(self.sample_x.shape[0]),self.sample_y,assume_a = 'sym')
        if get_timing == True:
            end = time.time()
            solve_time = end-start
        self.pred_mode = True
        print('Training finished.')
        if get_timing == True:
            print(f'Training time:{training_time}, solve_time: {solve_time}')


    def hgm_fast_pred(self,test_x,rtol = 1e-10, atol = 1e-10,debug_mode = False):
        start = time.time()
        dim = test_x.shape[1]
        X,Y = np.meshgrid(self.sample_x,test_x)
        pred_pts = np.hstack((X.reshape(test_x.shape[0]*self.sample_x.shape[0],1),Y.reshape(test_x.shape[0]*self.sample_x.shape[0],1)))
        pred_pts = batch_dl.add_index(pred_pts)
        
        
        sigma_array,sigma_dot_array = self.hgm_fast_ev_kernel(pred_pts,rtol = rtol, atol = atol,debug_mode = debug_mode)
        
        kernel_list = np.zeros(pred_pts.shape[0])

        
        for i in range(pred_pts.shape[0]):
            ntk = 0
           
            for s in range(self.layers):

                sigma_dot = 1
                for t in range(s+1,self.layers):
                    #print(sigma_dot_array[t-1][i])
                    sigma_dot = sigma_dot*sigma_dot_array[t-1][i]
                
                ntk = ntk +(sigma_dot*sigma_array[s][i])
            kernel_list[i] = ntk
        if debug_mode == True:
            print(kernel_list)
        kernel_list = kernel_list.reshape(-1,self.sample_x.shape[0])
       
        
        # print(kernel_list.shape)
        
        output = kernel_list@self.pred_matrix
        output = output.reshape(-1)
        print(output.shape)
        print(output)
        end = time.time()
        print('pred time:', end-start)
        return output
